# coding: utf-8
u"""Приложение для логирования запросов СМЭВ."""
import traceback
from cStringIO import StringIO

from spyne.model import Fault
from spyne.protocol import ProtocolBase
from spyne.server.wsgi import WsgiApplication
from spyne_smev.server.django import DjangoApplication

from educommon.ws_log.models import SmevLog


class LoggingDjangoApplication(DjangoApplication):
    u"""Переопределенный класс для логирования запросов СМЭВ."""

    def __init__(self, app, chunked=True, max_content_length=10 * 1024 * 1024,
                 block_length=8 * 1024):
        u"""Метод инициализации приложения для логирования запросов СМЭВ."""
        super(LoggingDjangoApplication, self).__init__(
            app, chunked, max_content_length, block_length)
        self.app.event_manager.add_listener('method_call', method_call)

    def generate_contexts(self, ctx, in_string_charset=None):
        u"""Call create_in_document and decompose_incoming_envelope.

        To get method_request string in order to generate contexts.
        """
        try:
            # sets ctx.in_document
            self.app.in_protocol.create_in_document(ctx, in_string_charset)

            # sets ctx.in_body_doc, ctx.in_header_doc and
            # ctx.method_request_string
            self.app.in_protocol.decompose_incoming_envelope(
                ctx, ProtocolBase.REQUEST)

            # returns a list of contexts. multiple contexts can be returned
            # when the requested method also has bound auxiliary methods.
            retval = self.app.in_protocol.generate_method_contexts(ctx)

        except Fault as e:
            log_object = ctx.transport.req['log']
            log_object.error = unicode(traceback.format_exc(), errors='ignore')
            ctx.in_object = None
            ctx.in_error = e
            ctx.out_error = e
            retval = (ctx,)

        return retval

    def __call__(self, request):
        u"""Логируем запрос, при вызове объекта приложения как функции."""
        request_body = request.META['wsgi.input'].read(
            request._stream.remaining)
        request.META['wsgi.input'] = StringIO(request_body)
        retval = self.HttpResponseObject()

        environ = request.META.copy()

        log_object = SmevLog()
        environ['log'] = log_object
        log_object.request = request_body

        def start_response(status, headers):
            # Status is one of spyne.const.http
            status, reason = status.split(' ', 1)

            retval.status_code = int(status)
            for header, value in headers:
                retval[header] = value

        try:
            response = WsgiApplication.__call__(self, environ, start_response)
            self.set_response(retval, response)
        except Exception:
            log_object.error = unicode(traceback.format_exc(), errors="ignore")
        else:
            log_object.response = retval.content
        finally:
            if request_body:
                log_object.service_address = request.build_absolute_uri()
                log_object.save()

        return retval


def method_call(ctx):
    u"""Обработчик события "Вызов метода"."""
    log = ctx.transport.req['log']

    # Добавляем имя метода в подготавливаемый лог-объект.
    log.method_name = ctx.method_name
    # Словарь METHOD_VERBOSE_NAMES обязательно должен быть определен в классе
    # приложения web-сервиса. Это словарь с ключами - имена сервисов и
    # значениями - кортежи с именем сервиса на русском языке и типом
    # взаимодействия.
    method_info = ctx.service_class.METHOD_VERBOSE_NAMES[ctx.method_name]
    log.method_verbose_name, log.interaction_type = method_info
